#!/usr/bin/perl -w

# Copyright 2021 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  See the file COPYING.  If not, see
# <http://www.gnu.org/licenses/>.

use strict;
use 5.006;
use FLAT;
use Test::More;

use FindBin;
use lib $FindBin::Bin;
use MyAppendMorphism ':all';
use MyFLAT;
use FomaInteractive;

use MyTestHelpers;
BEGIN { MyTestHelpers::nowarnings(); }

# uncomment this to run the ### lines
# use Smart::Comments;

plan tests => 46;


#------------------------------------------------------------------------------
# Morphism State Reversals

{
  my $morphism = MyAppendMorphism::HtoL_FLAT(1,1);
  # $morphism->MyFLAT::minimize(verbose=>1)->MyFLAT::view;

  foreach my $a (2) {
    my $r = MyAppendMorphism::HtoL_FLAT($a,1)
      ->MyFLAT::reverse
      ->MyFLAT::minimize(verbose=>1)
      ->MyFLAT::as_nfa;

    { my ($rstart) = $r->get_starting;
      my $next_symbol = ($a-1)%2;
      my ($rnext) = $r->successors($rstart, $next_symbol);
      $r->unset_starting($r->get_states);
      $r->set_starting($rnext);
      # $r->MyFLAT::view;
    }

    $r = $r->MyFLAT::minimize(verbose=>1)
      ->MyFLAT::as_nfa;
    is($r->num_states,4);

    my $num_loops = 0;
    foreach my $state ($r->get_states) {
      my ($loop) = $r->successors($state, 0);
      ### $loop
      if ($loop == $state) {
        $num_loops++;
        $r->unset_starting($r->get_states);
        $r->set_starting($state);

        $r->unset_accepting($r->get_accepting);
        $r->set_accepting($state);
        $r->MyFLAT::set_name("a=$a loop state=$state");

        my $d = $r->as_dfa;
        ok($d->equals($morphism),
           "a=$a loop state=$state");
        MyFLAT::FLAT_check_is_equal($d,$morphism);
      }
    }
    is($num_loops,2);
  }
}

# sub DFA_set_starting {
#   my ($fa,$starting) = @_;
#   $fa = $fa->MyFLAT::as_nfa($fa);
#   $fa->unset_starting($fa->get_starting);
#   $fa->set_starting($starting);
#   return $fa->as_dfa;
# }


#------------------------------------------------------------------------------
ok (defined $LtoH_starting_state);
ok ($LtoH_starting_state <= $LtoH_max_state);

#------------------------------------------------------------------------------
# LtoH

sub mod_14_23 {
  my ($str) = @_;
  $str =~ tr/14/23/;
  return $str;
}
{
  # equivalence of upper and lower 6x2 halves
  foreach my $s (1,4) {
    my $t = mod_14_23($s);
    foreach my $letter (qw(a b c d e f)) {
      my $s_name = "$s$letter";
      my $t_name = "$t$letter";
      my $s_state = $LtoH_name_to_state{$s_name} || die $s_name;
      my $t_state = $LtoH_name_to_state{$t_name} || die $t_name;
      foreach my $bit (0,1) {
        my $s_to = $LtoH_state_table[$s_state]->[$bit];
        my $t_to = $LtoH_state_table[$t_state]->[$bit];
        my $s_to_name = $LtoH_state_to_name[$s_to];
        my $t_to_name = $LtoH_state_to_name[$t_to];
        is ($t_to_name, mod_14_23($s_to_name));
      }
    }
  }
}


#------------------------------------------------------------------------------
# HtoL_FLAT() and LtoH_FLAT()

foreach my $a (1 .. 4) {
  my $htol = HtoL_FLAT($a);
  my $ltoh = LtoH_FLAT($a);
  ok (! $htol->contains(''));
  ok (! $ltoh->contains(''));

  # $htol accepts single 0 due to state sharing
  my $hrev = $htol->MyFLAT::reverse;
  ok ($hrev->equals($ltoh), "a=$a");
  # MyFLAT::FLAT_show_breadth($hrev, 4);
  # MyFLAT::FLAT_show_breadth($ltoh, 4);
  # MyFLAT::FLAT_check_is_equal($hrev,$ltoh);
}


#------------------------------------------------------------------------------

system ('perl gen-att.pl');

{
  # equivalence of transducers in Foma circa 0.9.18 may be false negatives,
  # and believe does so on HtoL-final.att versus LtoH-final.att
  #
  my $foma = FomaInteractive->new (verbose => 0);
  my $HtoL_filename = File::Spec->catdir($FindBin::Bin, "HtoL-final.att");
  my $LtoH_filename = File::Spec->catdir($FindBin::Bin, "LtoH-final.att");
  $foma->command ("read att $HtoL_filename",
                  "define HtoL",
                  "read att $LtoH_filename",
                  "define LtoH",
                  "regex HtoL;",
                  # "write att /tmp/x.att",
                  "regex LtoH;",
                  # "write att /tmp/y.att",
                 );
  # TODO: {
  #    local $TODO = 'Foma 0.9.18 false negatives on transducer equivalence';
  #    ok($foma->test('equivalent'),
  #       "HtoL-final.att equivalent LtoH-final.att");
  #  }
}


{
  foreach my $s (1 .. 4) {
    my $foma = FomaInteractive->new (verbose => 0);
    my $HtoL_filename = File::Spec->catdir($FindBin::Bin, "HtoL$s.att");
    my $LtoH_filename = File::Spec->catdir($FindBin::Bin, "LtoH$s.att");
    $foma->command ("define bit [\"0\"|1];",
                    "read att $HtoL_filename");
    $foma->command ("reverse net");
    $foma->command ("define hrev");

    $foma->command ("read att $LtoH_filename");
    $foma->command ("define LtoH");

    $foma->command ("regex hrev;",
                    "regex LtoH;");
    diag $foma->command ("print size");
    ok($foma->test('equivalent'),
       "s=$s equivalent");
  }

  # if ($s == 1) {
  #   $foma->command ("regex bit^>1;",
  #                   "intersect net");
  # }

  # $foma->command ("regex bit^<5;");
  # $foma->command ("intersect net");
  # $foma->command ("print words");

  # $foma->command ("regex bit^<5;",
  #                 "intersect net");
  # $foma->command ("print words");

  # $foma->command ("regex hrev - LtoH;");
  # $foma->command ("print words");
}

#------------------------------------------------------------------------------
exit 0;
